# doBackup.ps1 VERSION 1.0
# Copyright 2008, EJ innovations (ej@ejinnovations.com)
# 
# This script uses RoboCopy to make a backup of a source directory
# to a managed target directory.  This target directory will contain
# folders named after the date and times the backups were run. The 
# oldest backups will be deleted if necessary to make room for the next
# backup run.
#
# Usage:
#
#	DoBackup '\\someserver\somedirectory' 't:\'
#
# To Cap the amount of space used by the managed backup directory, 
# you can specify the gigabyte (or megabyte) ceiling as the last parameter.  If 
# omitted, the cap is the amount of free space on the drive.
# 	
# 	To set the backup limit to 9 GB...
# 	DoBackup '\\someserver\somedirectory' 't:\' '9G'
#
#

# Gets the disk usage of the specified path using
# the old shool FileSystemObject because it is a lot 
# faster than other methods.
function GetSize($path)
{
	# gotta use the scripting library to get the size quickly
	$fso = new-object -comobject Scripting.FileSystemObject;
	$folder = $fso.GetFolder($path.FullName);

	return $folder.Size;
}

# Get the amount of free space on a particular path's volume
# using WMI.  Note that Removable Media don't get their free space
# reported on Vista.
function GetFreeSpace($path)
{
	$root = (get-item $path).Root
	$deviceId = $root.Name.SubString(0,2);

	return (gwmi -query "select FreeSpace from win32_logicaldisk where deviceid='$deviceId'")["FreeSpace"]
}

# Prepares the managed backup directory to accomodate a new backup run
function PrepareSpace($sourceDirectory, $targetDirectory, $targetMaxSize)
{
	$maxBackupSize = GetTargetBackupMax $targetDirectory $targetMaxSize

	# get directoryentries for the two paths 
	$targetRoot = get-item $targetDirectory
	$source = get-item $sourceDirectory

	write-host "Calculating source size..."
	$sourceSize = GetSize $source
	write-host "Calculating target size...";
	$targetSize = GetSize $targetRoot

	$headroom = $maxBackupSize - $targetSize - $sourceSize

	if ($headroom -le 0)
	{
		# first see if the desired backup will even fit 
		if ($maxBackupSize -lt $sourceSize)
		{
			throw "Insufficient space for backup."
		}

		FreeSpace (-1 * $headroom) $targetRoot
	}
	else
	{
		write-host ($maxBackupSize - $targetSize) " bytes available for backup"
	}

	write-host $sourceSize " bytes in source"
}

# Deletes old backups to obtain param[0] more space free
function FreeSpace($extraSpaceNeeded, $root)
{
	$sortedDirList = new-object System.Collections.SortedList
	foreach ($subdirectory in $root.GetDirectories())
	{
		$sortedDirList.add($subdirectory.CreationTime, $subdirectory);
	}

	# go through each entry, oldest to newest
	$toDelete = @()
	foreach ($entry in $sortedDirList.GetEnumerator())
	{
		$subdirectory = $entry.Value
		$thisDirSize = GetSize $subdirectory

		$extraSpaceNeeded -= $thisDirSize;
		$toDelete = $toDelete + $subdirectory;

		# see if we need to continue looking for free space
		if ($extraSpaceNeeded -le 0)
		{
			break
		}
	}

	# safeguard
	if ($toDelete.Length -eq $sortedDirList.Length)
	{
		if (!$allowAllDelete)
		{
			throw "The only way to claim enough space for the backup is to delete all existing backups. Set allowAllDelete to True to continue."
		}
	}

	# actually do the delete
	foreach ($subdirectory in $toDelete)
	{
		write-host "Insufficient space for backup, deleting $subdirectory to claim additional space..."
		$subdirectory.Delete($TRUE)
	}
}

# Determines the maximum backup size we're allowed 
function GetTargetBackupMax($targetDirectory, $targetMaxSize)
{
	$targetSize = GetFreeSpace $targetDirectory

	if ($targetMaxSize.Length -gt 0)
	{
		# parse the specified size 
		if ($targetMaxSize.EndsWith("G"))
		{
			# interpret as Gigabytes
			if ($targetMaxSize -match "(\d+)")
			{
				$targetSize = [long]$matches[1] * 1024 * 1024 * 1024;
			}
		}
		else 
		{
			# interpret as Megabytes
			if ($targetMaxSize -match "(\d+)")
			{
				$targetSize = [long]$matches[1] * 1024 * 1024;
			}
		}
	}
	return $targetSize
}

# Send an email with the summary
function SendEmail($byteString, $timeString, $log)
{
	# this needs to be configured and uncommented
	#$SmtpClient = new-object system.net.mail.smtpClient
	#$SmtpClient.host = ""

	#$From = "PowerShell Backup <someone@somewhere.net>"
	#$To = "you@home.com"
	#$Title = "PowerShell Backup: " + $byteString
	#$Body = "Bytes: $byteString " + [environment]::newline + "Time: $timeString";
	#$SmtpClient.Send($from,$to,$title,$Body)  
}

# The main driver
function DoBackup
{
	param(
		[string]$sourceDirectory,
		[string]$targetDirectory,
		[string]$targetMaxSize = ""
	)
	PrepareSpace $sourceDirectory $targetDirectory $targetMaxSize

	$directoryName = [System.DateTime]::Now.ToString("yyyy.MM.dd H.mm.ss")
	$newDirectory = (get-item $targetDirectory).CreateSubDirectory($directoryName);

	write-host "Starting RoboCopy..."
	$roboOutput = robocopy "$sourceDirectory" "$newDirectory" /E /LOG:"$newDirectory"\robo.log /NP /TEE

	if ("$roboOutput" -match "Bytes : (.{20,20})")
	{
		$byteString = $matches[1]
	}
	else
	{
		$byteString = "FAIL"	
	}

	if ("$roboOutput" -match "Times : (.{20,20})")
	{
		$timeString = $matches[1]
	}
	else
	{
		$timeString = "FAIL";
	}

	write-host "Bytes: " $byteString
	write-host "Time: " $timeString

	SendEmail $byteString $timeString $roboOutput
}
